structure CDocs :> CDOCS =
  struct

  datatype cdoc =
    Txt of Context.ctxt * string
  | Ws  of Context.ctxt * string
  | Nd  of Context.ctxt * SMLNG.tag * cdocs
  and cdocs = One of cdoc | App of cdocs * cdocs

  fun fold (k as {kOne,kCons,kTxt,kWs,kNd}) cdocs = let
    fun f (Txt x) = kTxt x
      | f (Ws x) = kWs x
      | f (Nd (c,t,cds)) = kNd(c,t,g cds)
    and g (App(One x,xs)) = kCons(f x,g xs)
      | g (App(xs,ys)) = let
	  val kOne = (fn x => kCons(x,g ys))
	in fold {kOne=kOne,kCons=kCons,kTxt=kTxt,kWs=kWs,kNd=kNd} xs
	end 
      | g (One x) = kOne (f x)
  in g cdocs
  end

    fun fromDoc doc = Context.fold_with_ctxt 
      {kSeq=App,
       kTxt=(fn (c,s) => One (Txt(c,s))),
       kWs=(fn (c,s) => One (Ws(c,s))),
       kNd=(fn (c,t,a) => One (Nd(c,t,a)))} Context.init doc

    fun toDoc cdocs = 
      fold {kOne=(fn x => x),
	    kCons=SMLNG.Seq,
	    kTxt=(fn (c,s) => SMLNG.Txt s),
	    kWs=(fn (c,s) => SMLNG.Ws s),
	    kNd=(fn (c,t,ds) => SMLNG.Nd(t,ds))} cdocs

    fun ctxt (Txt(c,_)) = c
      | ctxt (Ws(c,_)) = c
      | ctxt (Nd(c,_,_)) = c

    fun tag (Nd(_,t,_)) = SOME t
      | tag _ = NONE 

  end
