structure Context :> CONTEXT =
  struct
    open SMLNG
    type ctxt = {b:bool,
		em:bool,
		 i:bool,
		 s:bool,
	        tt:bool,
		 u:int,
	      size:int option,
 	      color:color option}
    fun eq (c1:ctxt,c2:ctxt) = c1 = c2
    fun upd (B,{b,em,i,s,tt,u,size,color}) =
      {b=true,
       em=em,
       i=i,
       s=s,
       tt=tt,
       u=u,
       size=size,color=color}
      | upd (EM,{b,em,i,s,tt,u,size,color}) =
      {b=b,
       em=if not s then not em else em,
       i=i,
       s=s,
       tt=tt,
       u=u,
       size=size,color=color}
      | upd (I,{b,em,i,s,tt,u,size,color}) =
      {b=b,
       em=em,
       i=true,
       s=s,
       tt=tt,
       u=u,
       size=size,color=color}
      | upd (PL,{b,em,i,s,tt,u,size,color}) =
      {b=false,
       em=false,
       i=false,
       s=false,
       tt=tt,
       u=0,
       size=size,color=color}
      | upd (S,{b,em,i,s,tt,u,size,color}) =
      {b=b,
       em=false,
       i=i,
       s=true,
       tt=tt,
       u=u,
       size=size,color=color}
      | upd (TT,{b,em,i,s,tt,u,size,color}) =
      {b=b,
       em=em,
       i=i,
       s=s,
       tt=true,
       u=u,
       size=size,color=color}
      | upd (U,{b,em,i,s,tt,u,size,color}) =
      {b=b,
       em=em,
       i=i,
       s=s,
       tt=tt,
       u=if u < 3 then u + 1 else u,
       size=size,color=color}
      | upd (Size x,{b,em,i,s,tt,u,size,color}) =
      {b=b,
       em=em,
       i=i,
       s=s,
       tt=tt,
       u=u,
       size=SOME x,color=color}
      | upd (Color c,{b,em,i,s,tt,u,size,color}) =
      {b=b,
       em=em,
       i=i,
       s=s,
       tt=tt,
       u=u,
       size=size,color=SOME c}

    fun isSet (B,{b,em,i,s,tt,u,size,color}) = b = true
      | isSet (EM,{b,em,i,s,tt,u,size,color}) = s = true
      | isSet (I,{b,em,i,s,tt,u,size,color}) = i = true
      | isSet (PL,{b,em,i,s,tt,u,size,color}) =
      b = false  andalso em = false andalso  i = false  andalso
      s = false  andalso u = 0
      | isSet (S,{b,em,i,s,tt,u,size,color}) = s = true andalso em = false
      | isSet (TT,{b,em,i,s,tt,u,size,color}) = tt = true
      | isSet (U,{b,em,i,s,tt,u,size,color}) = u > 0
      | isSet (Size x,{b,em,i,s,tt,u,size,color}) = size = SOME x
      | isSet (Color c,{b,em,i,s,tt,u,size,color}) = color = SOME c

    val init =
      {b=false,em=false,i=false,s=false,tt=false,u=0,size=NONE,color=NONE}

    fun upd_ws {b,em,i,s,tt,u,size,color} =
      {b=false,em=false,i=false,s=false,tt=false,u=u,size=NONE,
       color=if u = 0 then NONE else color}

    fun b ({b=x,...}:ctxt) = x
    fun em ({em=x,...}:ctxt) = x
    fun i ({i=x,...}:ctxt) = x
    fun s ({s=x,...}:ctxt) = x
    fun tt ({tt=x,...}:ctxt) = x
    fun u ({u=x,...}:ctxt) = x
    fun size ({size=x,...}:ctxt) = x
    fun color ({color=x,...}:ctxt) = x

    fun fold_with_ctxt {kSeq,kTxt,kWs,kNd} ctxt d = let
      fun f (ctxt,Seq(d1,d2)) = kSeq(f (ctxt,d1),f (ctxt,d2))
	| f (ctxt,Txt s) = kTxt (ctxt,s)
	| f (ctxt,Ws s) = kWs (upd_ws ctxt,s)
	| f (ctxt,Nd (t,d)) = kNd(ctxt,t,f (upd(t,ctxt),d))
    in f(ctxt,d)
    end
    exception Bail
    fun common (x,y) = let
      fun sameB f = (f x) andalso (f y)
      fun sameO f =
	case (f x,f y) of
	  (NONE,NONE) => NONE
	| (SOME x,SOME y) => if x = y then SOME x else raise Bail
	| _ => raise Bail
    in (SOME {b=sameB b,
	     em=sameB em,
	     i=sameB i,
	     s=sameB s,
	     tt=sameB tt,
	     u=if (u x) = (u y) then u x else raise Bail,
	     size=sameO size,
	     color=sameO color}) handle Bail => NONE
    end

    fun toTags {b,em,i,s,tt,u,size,color} = let
      val tags = []
      val tags = if b  then B::tags else tags
      val tags = if em then EM::tags else tags
      val tags = if i  then I::tags else tags
      val tags = if s  then S::tags else tags
      val tags = if tt then TT::tags else tags
      val tags =
	case u of 
	  0 => tags
	| 1 => U::tags
	| 2 => U::U::tags
	| 3 => U::U::U::tags
	| _ => raise (Fail "bug")
      val tags =
	case size of
	  NONE => tags
	| SOME size => (Size size)::tags
      val tags =
	case color of
	  NONE => tags
	| SOME color => (Color color)::tags
    in tags
    end

  end